# -*- coding: utf-8 -*-
"""Module for certificates
"""
# import basic stuff
import logging

# import own stuff
from tlsmate import tls
from tlsmate import cert_utils
from tlsmate.exception import UntrustedCertificate

# import other stuff
from cryptography import x509
from cryptography.x509.oid import NameOID, ExtensionOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import (
    ec,
    rsa,
    dsa,
    ed25519,
    ed448,
)
from cryptography.hazmat.primitives.serialization import Encoding


# list of EV OIDs. Sources:
# - https://chromium.googlesource.com/chromium/src/net/+/refs/heads/main/cert/ev_root_ca_metadata.cc  # noqa
# - https://hg.mozilla.org/mozilla-central/file/tip/security/certverifier/ExtendedValidation.cpp  # noqa
_ev_oids = [
    "1.2.156.112559.1.1.6.1",
    "1.2.392.200091.100.721.1",
    "1.2.616.1.113527.2.5.1.1",
    "1.3.159.1.17.1",
    "1.3.171.1.1.10.5.2",
    "1.3.6.1.4.1.13177.10.1.3.10",
    "1.3.6.1.4.1.14777.6.1.1",
    "1.3.6.1.4.1.14777.6.1.2",
    "1.3.6.1.4.1.17326.10.14.2.1.2",
    "1.3.6.1.4.1.17326.10.14.2.2.2",
    "1.3.6.1.4.1.34697.2.1",
    "1.3.6.1.4.1.34697.2.2",
    "1.3.6.1.4.1.34697.2.3",
    "1.3.6.1.4.1.34697.2.4",
    "1.3.6.1.4.1.40869.1.1.22.3",
    "1.3.6.1.4.1.4146.1.1",
    "1.3.6.1.4.1.4788.2.202.1",
    "1.3.6.1.4.1.6334.1.100.1",
    "1.3.6.1.4.1.6449.1.2.1.5.1",
    "1.3.6.1.4.1.782.1.2.1.8.1",
    "1.3.6.1.4.1.7879.13.24.1",
    "1.3.6.1.4.1.8024.0.2.100.1.2",
    "2.16.156.112554.3",
    "2.16.528.1.1003.1.2.7",
    "2.16.578.1.26.1.3.3",
    "2.16.756.1.89.1.2.1.1",
    "2.16.756.5.14.7.4.8",
    "2.16.792.3.0.4.1.1.4",
    "2.16.840.1.114028.10.1.2",
    "2.16.840.1.114404.1.1.2.4.1",
    "2.16.840.1.114412.2.1",
    "2.16.840.1.114413.1.7.23.3",
    "2.16.840.1.114414.1.7.23.3",
    "2.16.840.1.114414.1.7.24.3",
    "2.23.140.1.1",
]


class Certificate(object):
    """Represents a certificate.

    The arguments der and pem are exclusive.

    Arguments:
        der (bytes): the certificate in DER-format (raw bytes)
        pem (bytes): the certificate in PEM-format
        parse (bool): whether the certificate shall be parsed (i.e., all relevant
            data are extracted from the given der/pem structure), or if it shall just
            be stored. In the latter case the certificate will be parsed if a
            property is accessed.
    """

    def __init__(self, der=None, pem=None, x509_cert=None, parse=False):

        if (der, pem, x509_cert).count(None) != 2:
            raise ValueError("der, pem and x509_cert are exclusive")

        self._bytes = None
        self._pem = None
        self._parsed = None
        self._subject_str = None
        self._self_signed = None
        self.auth_key_id = None
        self.subject_key_id = None
        self.subject_matches = None
        self.fingerprint_sha1 = None
        self.fingerprint_sha256 = None
        self.tls12_signature_algorithms = None
        self.tls13_signature_algorithms = None
        self.crl_status = None
        self.ocsp_status = None
        self.issues = []
        self.trusted = tls.ScanState.UNDETERMINED
        self.tls_extensions = []
        self.issuer_cert = None
        self.ocsp_must_staple = tls.ScanState.FALSE
        self.ocsp_must_staple_multi = tls.ScanState.FALSE
        self.extended_validation = tls.ScanState.NA
        self.from_trust_store = False

        if der is not None:
            self._bytes = der
            if parse:
                self._parsed = x509.load_der_x509_certificate(self._bytes)
                self._parse()

        elif pem is not None:
            if isinstance(pem, str):
                pem = pem.encode()

            self._pem = pem
            self._parsed = x509.load_pem_x509_certificate(pem)
            self._parse()

        else:
            self._parsed = x509_cert
            if parse:
                self._parse()

    def __str__(self):
        return self.subject_str

    def __eq__(self, other):
        return self.bytes == other.bytes

    @property
    def subject_str(self):
        """str: The subject name formatted according to RFC 4514.
        """
        if self._subject_str is None:
            self._subject_str = self._parsed.subject.rfc4514_string()
        return self._subject_str

    @property
    def parsed(self):
        """:obj:`cryptography.x509.Certificate`: the x509 certificate object
        """
        if self._parsed is None:
            self._parsed = x509.load_der_x509_certificate(self._bytes)
            self._parse()
        return self._parsed

    @property
    def bytes(self):
        """bytes: the certificate in raw format, i.e. in DER format.
        """
        if self._bytes is None:
            self._bytes = self.parsed.public_bytes(Encoding.DER)
        return self._bytes

    @property
    def pem(self):
        """bytes: the certificate in pem format (it is a binary string!)"""
        if self._pem is None:
            self._pem = self.parsed.public_bytes(Encoding.PEM)
        return self._pem

    def _determine_signature_algorithms(self, public_key):
        """For a given public key provide the compatible signature algorithms.
        """
        if isinstance(public_key, rsa.RSAPublicKey):
            self.tls12_signature_algorithms = [
                tls.SignatureScheme.RSA_PKCS1_SHA1,
                tls.SignatureScheme.RSA_PKCS1_SHA256,
                tls.SignatureScheme.RSA_PKCS1_SHA384,
                tls.SignatureScheme.RSA_PKCS1_SHA512,
                tls.SignatureScheme.RSA_PKCS1_MD5,
                tls.SignatureScheme.RSA_PKCS1_SHA224,
                # Currently, cryptography does not support RSA-PSS-PSS
                # tls.SignatureScheme.RSA_PSS_PSS_SHA256,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA384,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            ]
            self.tls13_signature_algorithms = [
                # tls.SignatureScheme.RSA_PSS_PSS_SHA256,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA384,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            ]

        elif isinstance(public_key, dsa.DSAPublicKey):
            self.tls12_signature_algorithms = [
                tls.SignatureScheme.DSA_MD5,
                tls.SignatureScheme.DSA_SHA1,
                tls.SignatureScheme.DSA_SHA224,
                tls.SignatureScheme.DSA_SHA256,
                tls.SignatureScheme.DSA_SHA384,
                tls.SignatureScheme.DSA_SHA512,
            ]
            self.tls13_signature_algorithms = []

        elif isinstance(public_key, ec.EllipticCurvePublicKey):
            size_to_algo = {
                128: tls.SignatureScheme.ECDSA_SHA1,
                224: tls.SignatureScheme.ECDSA_SECP224R1_SHA224,
                256: tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                384: tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                512: tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            }
            sig_scheme = size_to_algo.get(public_key.curve.key_size)
            if sig_scheme is None:
                raise ValueError(
                    f"unknown key size {public_key.curve.key_size} for ECDSA public key"
                )

            self.tls12_signature_algorithms = [sig_scheme]
            if sig_scheme is tls.SignatureScheme.ECDSA_SHA1:
                self.tls13_signature_algorithms = []
            else:
                self.tls13_signature_algorithms = [sig_scheme]

        elif isinstance(public_key, ed25519.Ed25519PublicKey):
            self.tls12_signature_algorithms = [tls.SignatureScheme.ED25519]
            self.tls13_signature_algorithms = [tls.SignatureScheme.ED25519]

        elif isinstance(public_key, ed448.Ed448PublicKey):
            self.tls12_signature_algorithms = [tls.SignatureScheme.ED448]
            self.tls13_signature_algorithms = [tls.SignatureScheme.ED448]

    def _parse(self):
        """Parse the certificate, so that all attributes are set.
        """
        self.fingerprint_sha1 = self._parsed.fingerprint(hashes.SHA1())
        self.fingerprint_sha256 = self._parsed.fingerprint(hashes.SHA256())
        self.signature_algorithm = cert_utils.map_x509_sig_scheme(
            self._parsed.signature_hash_algorithm, self._parsed.signature_algorithm_oid,
        )
        try:
            key_usage = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.KEY_USAGE
            )
            if key_usage.value.digital_signature:
                self._determine_signature_algorithms(self._parsed.public_key())
        except x509.ExtensionNotFound:
            self._determine_signature_algorithms(self._parsed.public_key())

        try:
            self.auth_key_id = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.AUTHORITY_KEY_IDENTIFIER
            ).value.key_identifier

        except x509.ExtensionNotFound:
            self.auth_key_id = None

        try:
            self.subject_key_id = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.SUBJECT_KEY_IDENTIFIER
            ).value.digest

        except x509.ExtensionNotFound:
            self.subject_key_id = None

        try:
            tls_features = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.TLS_FEATURE
            ).value

            if x509.TLSFeatureType.status_request in tls_features:
                self.ocsp_must_staple = tls.ScanState.TRUE

            if x509.TLSFeatureType.status_request_v2 in tls_features:
                self.ocsp_must_staple_multi = tls.ScanState.TRUE

        except x509.ExtensionNotFound:
            pass

        try:
            basic_constr = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.BASIC_CONSTRAINTS
            ).value
            if not basic_constr.ca:
                try:
                    cert_policies = self._parsed.extensions.get_extension_for_oid(
                        ExtensionOID.CERTIFICATE_POLICIES
                    ).value

                    if any(
                        pol.policy_identifier.dotted_string in _ev_oids
                        for pol in cert_policies
                    ):
                        self.extended_validation = tls.ScanState.TRUE

                    else:
                        self.extended_validation = tls.ScanState.FALSE

                except x509.ExtensionNotFound:
                    pass

        except x509.ExtensionNotFound:
            pass

    def _common_name(self, name):
        """From a given name, extract the common name

        Note, that there might be multiple common names present, in this case
        simple the first one is returned.
        """
        cns = name.get_attributes_for_oid(NameOID.COMMON_NAME)
        if not cns:
            raise UntrustedCertificate(f'no common name for "{self}"')
        return cns[0].value

    @property
    def self_signed(self):
        """bool: Provide an indication if the certificate is self-signed.
        """
        if self._self_signed is None:
            self._self_signed = cert_utils.equal_names(
                self.parsed.subject, self.parsed.issuer
            )
        return self._self_signed

    def mark_untrusted(self, issue):
        """Mark the certificate as untrusted.

        Arguments:
            issue (str): the error message containing the reason

        Raises:
            :obj:`tlsmate.exception.UntrustedCertificate`: if raise_on_failure is True
        """

        self.trusted = tls.ScanState.FALSE
        issue_long = f"certificate {self}: {issue}"
        logging.debug(issue_long)
        self.issues.append(issue)

    def validate_period(self, datetime, raise_on_failure=True):
        """Validate the period of the certificate against a given timestamp.

        Arguments:
            datetime (:obj:`datetime.datetime`): the timestamp
            raise_on_failure (bool): whether an exception shall be raised if the
                validation fails.

        Returns:
            bool: The validation state

        Raises:
            :obj:`tlsmate.exception.UntrustedCertificate`: if the timestamp is
                outside the validity period and raise_on_failure is True
        """

        if datetime < self.parsed.not_valid_before:
            self.mark_untrusted("validity period not yet reached", raise_on_failure)

        if datetime > self.parsed.not_valid_after:
            self.mark_untrusted("validity period exceeded", raise_on_failure)

    def has_valid_period(self, timestamp):
        """Determines if the period is valid.

        Arguments:
            timestamp (datetime.datetime): the timestamp to check against

        Returns:
            bool: An indication if the period is valid
        """
        valid = True
        if timestamp < self.parsed.not_valid_before:
            self.mark_untrusted("validity period not yet reached")
            valid = False

        if timestamp > self.parsed.not_valid_after:
            self.mark_untrusted("validity period exceeded")
            valid = False

        return valid

    def has_valid_subject(self, domain):
        """Validate if the certificate matches the given domain

        It takes the subject and the subject alternative name into account, and
        supports wildcards as well.

        Arguments:
            domain (str): the domain to check against (normally used in the SNI)
            raise_on_failure (bool): whether an exception shall be raised if the
                validation fails or not. Useful for a TLS scan, as the scan shall
                continue.

        Returns:
            bool: indication, if the domain name matches the certificate's subject/SAN

        Raises:
            UntrustedCertificate: if the certificate is not issued for the given
            domain
        """

        subject_matches = False
        domain = cert_utils.string_prep(domain)
        no_subdomain = cert_utils.remove_subdomain(domain)

        subject_cn = self._common_name(self.parsed.subject)
        if cert_utils.subject_matches(subject_cn, domain, no_subdomain):
            subject_matches = True

        else:
            try:
                subj_alt_names = self.parsed.extensions.get_extension_for_oid(
                    ExtensionOID.SUBJECT_ALTERNATIVE_NAME
                )
                for name in subj_alt_names.value.get_values_for_type(x509.DNSName):
                    if cert_utils.subject_matches(name, domain, no_subdomain):
                        subject_matches = True
                        break

            except x509.ExtensionNotFound:
                pass

        if not subject_matches:
            self.mark_untrusted("subject name does not match")

        self.subject_matches = subject_matches
        return subject_matches

    def validate_signature(self, sig_scheme, data, signature):
        """Validate a signature using the public key from the certificate.

        Arguments:
            sig_scheme (:class:`tlsmate.tls.SignatureScheme`): The signature
                scheme to use
            data (bytes): the bytes for which the signature is to be validated
            signature (bytes): the signature

        Raises:
            cryptography.exceptions.InvalidSignature: If the signature does not
                validate.

        """
        cert_utils.validate_signature(self, sig_scheme, data, signature)

    def validate_cert_signature(self, cert):
        """Validate the signature within a certificate

        Arguments:
            cert (:obj:`Certificate`): the certificate for which the signature
                shall be checked.

        Raises:
            cryptography.exceptions.InvalidSignature: If the signature does not
                validate.
        """

        self.validate_signature(
            cert.signature_algorithm,
            cert.parsed.tbs_certificate_bytes,
            cert.parsed.signature,
        )



# -*- coding: utf-8 -*-
"""Module defining various structures
"""
# import basic stuff
from typing import NamedTuple, Any

# import own stuff
from tlsmate import tls
from tlsmate.kdf import Kdf

# import other stuff
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding


class SessionStateId(NamedTuple):
    """Set of items to store a session id in the client.
    """

    session_id: bytes
    cipher_suite: tls.CipherSuite
    version: tls.Version
    master_secret: bytes


class SessionStateTicket(NamedTuple):
    """Set of items to store a session ticket in the client.
    """

    ticket: bytes
    lifetime: int
    cipher_suite: tls.CipherSuite
    version: tls.Version
    master_secret: bytes


class EarlyData(NamedTuple):
    """Set of items required to determine the early data traffic keys
    """

    kdf: Kdf
    early_secret: bytes
    mac_len: int
    binders_bytes: bytes = None


class Cipher(NamedTuple):
    """Set of properties describing a cipher.
    """

    primitive: tls.CipherPrimitive = None
    algo: algorithms.AES = None
    c_type: tls.CipherType = None
    key_len: int = None
    block_size: int = None
    iv_len: int = None
    tag_length: int = None
    cipher_supported: bool = False


class Mac(NamedTuple):
    """Set of properties describing a MAC.
    """

    hash_algo: hashes.SHA1  # just an example
    mac_len: int
    key_len: int
    hmac_algo: hashes.SHA1  # just an example


class Psk(NamedTuple):
    """Set of items to store a TLS1.3 pre-shared key in the client.
    """

    psk: bytearray
    lifetime: int
    age_add: int
    ticket: bytes
    timestamp: float
    cipher_suite: tls.CipherSuite
    version: tls.Version
    hmac: Mac


class SymmetricKeys(NamedTuple):
    """Set of keys
    """

    mac: bytes
    enc: bytes
    iv: bytes


class StateUpdateParams(NamedTuple):
    """Set of properties used to update the record layer state.
    """

    cipher: Cipher
    mac: Mac
    keys: SymmetricKeys
    compr: tls.CompressionMethod
    enc_then_mac: bool
    version: tls.Version
    is_write_state: bool


class CipherSuite(NamedTuple):
    """Set of properties for a cipher suite.
    """

    key_ex: tls.KeyExchangeAlgorithm
    cipher: tls.SymmetricCipher
    mac: tls.HashPrimitive


class RecordLayerMsg(NamedTuple):
    """Set of properties describing a record layer message.
    """

    content_type: tls.ContentType
    version: tls.Version
    fragment: bytes


class UpperLayerMsg(NamedTuple):
    """Set of properties describing an upper layer message.

    "Upper Layer" comprises handshake, ccs and alert protocols.
    """

    content_type: tls.ContentType
    msg_type: tls.HandshakeType  # can be CCSType as well
    msg: bytes


class KeyExchange(NamedTuple):
    """Set of properties describing a key exchange method.
    """

    key_ex_type: tls.KeyExchangeType = None
    key_auth: tls.KeyAuthentication = None
    key_ex_supported: bool = False
    default_sig_scheme: tls.SignatureScheme = None


class KeyShareEntry(NamedTuple):
    """Set of properties describing a key share entry.
    """

    group: tls.SupportedGroups
    key_exchange: bytes


class CipherSuiteDetails(NamedTuple):
    """Structure which provides details for a cipher suite.
    """

    cipher_suite: tls.CipherSuite
    full_hs: bool = False
    key_exchange_supported: bool = False
    key_algo: tls.KeyExchangeAlgorithm = None
    key_algo_struct: KeyExchange = None
    cipher: tls.SymmetricCipher = None
    cipher_struct: Cipher = None
    mac: tls.HashPrimitive = None
    mac_struct: Mac = None


class ProfileValues(NamedTuple):
    """Structure for the most relevant parameters for client hello
    """

    versions: list = []
    cipher_suites: list = []
    supported_groups: list = []
    signature_algorithms: list = []
    key_shares: list = []
    """Only applicable for TLS1.3"""


class CertSigAlgo(NamedTuple):
    """Structure for certificate signature algorithms
    """

    algo: hashes.SHA1 = None
    padd: padding.PKCS1v15 = None


class TransportEndpoint(NamedTuple):
    """Structure for host, type and port
    """

    host: str
    port: int
    host_type: tls.HostType


class ResolvedHost(NamedTuple):
    """Result of a DNS resolution
    """

    ipv4_addresses: list
    ipv6_addresses: list


class ConfigItem(NamedTuple):
    """A configuration setting
    """

    name: str
    default: Any = None
    type: Any = str


class Malfunction(NamedTuple):
    """Structure for server malfunction
    """

    issue: tls.ServerIssue
    message: tls.HandshakeType = None
    extension: tls.Extension = None



# -*- coding: utf-8 -*-
"""Module containing the CLI implementation
"""
# import basic stuff
import importlib
import pkgutil
import sys

# import own stuff
from tlsmate.config import Configuration
from tlsmate.tlsmate import TlsMate
from tlsmate.plugin import BaseCommand
from tlsmate import utils
from tlsmate.plugin import WorkManager

# import other stuff


def build_parser():
    """Creates the parser object

    Returns:
        :obj:`argparse.ArgumentParser`: the parser object as created with argparse
    """

    return BaseCommand.create_parser()


def main():
    """The entry point for the command line interface
    """

    utils.set_logging_format()

    parser = build_parser()
    args = parser.parse_args()

    # logging should be setup as early as possible
    utils.set_logging_level(args.logging)

    config = Configuration()
    BaseCommand.register_config(config)
    config.init_from_external(args.config_file)
    config.set("logging", args.logging)
    work_manager = WorkManager()
    BaseCommand.args_parsed(args, parser, None, config)
    tlsmate = TlsMate(config=config)
    work_manager.run(tlsmate)


# And now load the plugins which are shipped by default with tlsmate...
from tlsmate.plugins import scan, version  # NOQA

# And now look for additional user provided plugins
if len(sys.argv) < 2 or sys.argv[1] != "--no-plugin":
    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith("tlsmate_"):
            importlib.import_module(name)



# -*- coding: utf-8 -*-
"""Module containing the class for the configuration
"""
# import basic stuff
import os
import logging
from pathlib import Path

# import own stuff
from tlsmate.structs import ConfigItem

# import other stuff
import configparser

config_logging = ConfigItem("logging", type=str, default="error")
config_host = ConfigItem("host", type=str, default="localhost")
config_port = ConfigItem("port", type=int, default=443)
config_interval = ConfigItem("interval", type=int, default=0)
config_key_log_file = ConfigItem("key_log_file", type=str, default=None)
config_progress = ConfigItem("progress", type=bool, default=False)
config_sni = ConfigItem("sni", type=str, default=None)
config_ca_certs = ConfigItem("ca_certs", type="file_list")
config_client_key = ConfigItem("client_key", type="file_list")
config_client_chain = ConfigItem("client_chain", type="file_list")
config_crl = ConfigItem("crl", type=bool, default=True)
config_ocsp = ConfigItem("ocsp", type=bool, default=True)


class Configuration(object):
    """Class representing the configuration for tlsmate.

    The configuration is taken from the following sources (the list is ordered
    according to the priority, first item has the least priority):

    * The hard coded default values
    * From the ini-file, if present. If the ini-file is not specified via the
      CLI option, the file .tlsmate.ini in the user's home directory will be used,
      if present.
    * Environment variables. They need to be given in upper cases and must start with
      `TLSMATE_` followed by the name of the setting in upper cases.
    * From the command line interface parameters

    Example:
        Specifying the logging option:

        Via command line::

            tlsmate --logging=debug ...

        Via Enviroment variable::

            export TLSMATE_LOGGING=debug

        Via ini-file::

            [tlsmate]
            logging = debug
    """

    def __init__(self):
        self._config = {}
        self._descr = {}

        # register configurations which are essential in the core
        # part of tlsmate.
        for item in [
            config_logging,
            config_host,
            config_port,
            config_interval,
            config_key_log_file,
            config_progress,
            config_sni,
            config_ca_certs,
            config_client_key,
            config_client_chain,
            config_crl,
            config_ocsp,
        ]:
            self.register(item)

        # special configuration item exclusively used for unit tests.
        # if set to False, the recorder will not use any delays when replaying.
        self.register(ConfigItem("recorder_delay", type=bool, default=True))
        self._init_environment_var("recorder_delay")

    def _str_to_filelist(self, string):
        """Resolves a string of files paths.

        Multiple file paths are separated by a colon. If the path is a relative
        path, it is expanded to an absolute path, taking the directory of the
        config file as a base (which is available in self._config_dir).

        Arguments:
            string (str): the list of file paths as a string, multiple paths are
                separated by a colon.

        Returns:
            list of str: the list of resolved absolute paths.
        """

        ret = []
        for val in string.split(","):
            val = val.strip()
            if not val.startswith("/"):
                val = str(self._config_dir / val)
            ret.append(val)
        return ret

    _format_option = {
        bool: lambda self, x: x.lower() not in ["0", "off", "no", "false"],
        int: lambda self, x: int(x),
        "file_list": _str_to_filelist,
    }

    def _init_from_ini_file(self, ini_file):
        """Helper method to initialize the configuration from an ini-file.
        """

        if ini_file is None:
            ini_file = Path.home() / ".tlsmate.ini"
            if not ini_file.is_file():
                return

        else:
            ini_file = Path(ini_file)
            if not ini_file.is_absolute():
                ini_file = Path.cwd() / ini_file

        if not ini_file.is_file():
            raise FileNotFoundError(ini_file)

        logging.debug(f"using config file {str(ini_file)}")
        self._config_dir = ini_file.parent
        parser = configparser.ConfigParser()
        parser.read(str(ini_file))
        if parser.has_section("tlsmate"):
            config = parser["tlsmate"]
            for item in self._config:
                val = config.get(item)
                if val is not None:
                    self._config[item] = self._cast_item(item, val)

    def _init_environment_var(self, item):
        """Helper method to initialize a single item from the environment variable.
        """

        val = os.environ.get("TLSMATE_" + item.upper())
        if val is not None:
            self._config[item] = self._cast_item(item, val)

    def _init_from_environment(self):
        """Helper method to initialize the configuration from environment variables.
        """

        for item in self._config:
            self._init_environment_var(item)

    def _cast_item(self, item, val):
        """Cast the type of a configuration item into the internal format.
        """

        item_type = self._descr[item].type
        if item_type in self._format_option:
            val = self._format_option[item_type](self, val)
        return val

    def init_from_external(self, ini_file=None):
        """Take the configuration from the ini file and from the environment variables.

        Arguments:
            ini_file (str): the path to the ini file. If None is given, then only
                the environment variables are taken into account.
        """

        self._init_from_ini_file(ini_file)
        self._init_from_environment()

    def items(self):
        """Return the configuration items. Mimics the dict's `items` method.

        Returns:
            list (tuple): The list of configuration items. Each item is a tuple of the
            item name and the item value.

        """

        return self._config.items()

    def get(self, key, default=None):
        """Get the value of a configuration item. Mimics the dict's `get` method.

        Arguments:
            key (str): the name of the configuration item
            default: the default value to return in case the configuration item is
                not existing. Defaults to None.

        Returns:
            any: the value of the configuration item or the provided default value if
            it is not present.
        """

        return self._config.get(key, default)

    def set(self, key, val, keep_existing=True):
        """Add a configuration option.

        Arguments:
            key (str): the name of the option
            val: the value of the option
            keep_existing (bool): if set to True and the value is None, an existing
                configuration will not be overwritten. Defaults to True
        """

        if key in self._config and val is None and keep_existing:
            return

        self._config[key] = val

    def register(self, config_item):
        """Add a new item to the configuration

        Arguments:
            config_item (:obj:`tlsmate.structs.ConfigItem`): the configuration
                item to register

        Raises:
            ValueError: if a configuration item with the same name is already existing
        """

        name = config_item.name
        if name in self._descr:
            if config_item != self._descr[name]:
                raise ValueError(
                    f'configuration setting "{name}" already defined with '
                    f"different properties"
                )
        self._descr[name] = config_item
        self._config[name] = config_item.default



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.supported_groups import ScanSupportedGroups
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_basic_openssl1_0_2_DSA"
    sp_out_yaml = "profile_supported_groups_openssl1_0_2_DSA"
    recorder_yaml = "recorder_supported_groups_openssl1_0_2_DSA"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-dsa "
        "-- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_2

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        ScanSupportedGroups(tlsmate).run()

        profile = server_profile.make_serializable()

        assert "supported_groups" not in profile["versions"][0]
        assert len(profile["versions"][1]["supported_groups"]) == 0
        assert len(profile["versions"][2]["supported_groups"]) == 0
        assert len(profile["versions"][3]["supported_groups"]) == 0
        assert len(profile["versions"][4]["supported_groups"]) == 0
        assert "supported_groups" not in profile["versions"][5]


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module for the version plugin
"""
# import basic stuff

# import own stuff
from tlsmate.plugin import BaseCommand, Plugin, Args, Worker
from tlsmate.version import __version__

# import other stuff


class VersionWorker(Worker):
    """Worker to print the version of tlsmate.
    """

    name = "version"

    def run(self):
        print(__version__)


@BaseCommand.extend
class SubcommandVersion(Plugin):
    """CLI plugin to print the version of ``tlsmate``.
    """

    subcommand = Args("version", help="prints the version of tlsmate")
    workers = [VersionWorker]



# -*- coding: utf-8 -*-
"""Module containing classes for the key exchange
"""
# import basic stuff
import abc
from typing import NamedTuple
import os

# import own stuff
from tlsmate import dh_numbers
from tlsmate import tls
from tlsmate import pdu
from tlsmate.exception import ServerMalfunction, CurveNotSupportedError
from tlsmate import kdf

# import other stuff
from cryptography.hazmat.primitives.asymmetric import ec, x25519, x448, dh
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    PrivateFormat,
    NoEncryption,
    load_der_private_key,
)


def verify_signed_params(prefix, params, cert, default_scheme, version):
    """Verify the signed parameters from a ServerKeyExchange message.

    Arguments:
        prefix (bytes): the bytes to prepend to the data
        params: the parameter block from the ServerKeyExchange message
        cert (:obj:`tlsmate.cert.Certificate`): the certificate used to validate
            the data
        default_scheme (:class:`tlsmate.tls.SignatureScheme`): the default
            signature scheme to use (if not present in the message)
        version (:class:`tlsmate.tls.Version`): the TLS version. For TLS1.1 and below
            the signature is constructed differently (using SHA1 + MD digests)
    Raises:
        cryptography.exceptions.InvalidSignature: If the signature does not
            validate.
    """

    data = prefix + params.signed_params

    if (
        default_scheme is tls.SignatureScheme.RSA_PKCS1_SHA1
        and version is not tls.Version.TLS12
    ):
        # Digest is a combination of MD5 and SHA1
        digest = kdf.Kdf()
        digest.start_msg_digest()
        digest.set_msg_digest_algo(None)
        digest.update_msg_digest(data)
        hashed1 = digest.current_msg_digest(suspend=True)
        key = cert.parsed.public_key()
        hashed2 = key.recover_data_from_signature(
            params.signature, padding.PKCS1v15(), None
        )
        if hashed1 != hashed2:
            raise InvalidSignature

    else:
        sig_scheme = params.sig_scheme or default_scheme
        cert.validate_signature(sig_scheme, data, params.signature)


def verify_certificate_verify(cert_ver, msgs, msg_digest):
    """Validates the CertificateVerify message.

    Arguments:
        cert_ver (:obj:`tlsmate.msg.CertificateVerify`): the CertificateVerify
            message to verify
        msgs (:obj:`tlsmate.connection.TlsConnectionMsgs`): the object storing
            received/sent messages
        msg_digest (bytes): the message digest used to construct the data to validate

    Raises:
        cryptography.exceptions.InvalidSignature: If the signature does not
            validate.
    """

    data = (b" " * 64) + b"TLS 1.3, server CertificateVerify" + b"\0" + msg_digest
    cert = msgs.server_certificate.chain.certificates[0]
    cert.validate_signature(cert_ver.signature_scheme, data, cert_ver.signature)


def _verify_rsa_pkcs(pub_key, signature, data, hash_algo):
    pub_key.verify(signature, data, padding.PKCS1v15(), hash_algo())


def _verify_dsa(pub_key, signature, data, hash_algo):
    pub_key.verify(signature, data, hash_algo())


def _verify_ecdsa(pub_key, signature, data, hash_algo):
    pub_key.verify(signature, data, ec.ECDSA(hash_algo()))


def _verify_xcurve(pub_key, signature, data, hash_algo):
    pub_key.verify(signature, data)


def _verify_rsae_pss(pub_key, signature, data, hash_algo):
    pub_key.verify(
        signature,
        data,
        padding.PSS(mgf=padding.MGF1(hash_algo()), salt_length=hash_algo.digest_size),
        hash_algo(),
    )


_sig_schemes = {
    tls.SignatureScheme.RSA_PKCS1_MD5: (_verify_rsa_pkcs, hashes.MD5),
    tls.SignatureScheme.RSA_PKCS1_SHA1: (_verify_rsa_pkcs, hashes.SHA1),
    tls.SignatureScheme.RSA_PKCS1_SHA224: (_verify_rsa_pkcs, hashes.SHA224),
    tls.SignatureScheme.RSA_PKCS1_SHA256: (_verify_rsa_pkcs, hashes.SHA256),
    tls.SignatureScheme.RSA_PKCS1_SHA384: (_verify_rsa_pkcs, hashes.SHA384),
    tls.SignatureScheme.RSA_PKCS1_SHA512: (_verify_rsa_pkcs, hashes.SHA512),
    tls.SignatureScheme.DSA_MD5: (_verify_dsa, hashes.MD5),
    tls.SignatureScheme.DSA_SHA1: (_verify_dsa, hashes.SHA1),
    tls.SignatureScheme.DSA_SHA224: (_verify_dsa, hashes.SHA224),
    tls.SignatureScheme.DSA_SHA256: (_verify_dsa, hashes.SHA256),
    tls.SignatureScheme.DSA_SHA384: (_verify_dsa, hashes.SHA384),
    tls.SignatureScheme.DSA_SHA512: (_verify_dsa, hashes.SHA512),
    tls.SignatureScheme.ECDSA_SHA1: (_verify_ecdsa, hashes.SHA1),
    tls.SignatureScheme.ECDSA_SECP256R1_SHA256: (_verify_ecdsa, hashes.SHA256),
    tls.SignatureScheme.ECDSA_SECP384R1_SHA384: (_verify_ecdsa, hashes.SHA384),
    tls.SignatureScheme.ECDSA_SECP521R1_SHA512: (_verify_ecdsa, hashes.SHA512),
    tls.SignatureScheme.RSA_PSS_RSAE_SHA256: (_verify_rsae_pss, hashes.SHA256),
    tls.SignatureScheme.RSA_PSS_RSAE_SHA384: (_verify_rsae_pss, hashes.SHA384),
    tls.SignatureScheme.RSA_PSS_RSAE_SHA512: (_verify_rsae_pss, hashes.SHA512),
    tls.SignatureScheme.ED25519: (_verify_xcurve, None),
    tls.SignatureScheme.ED448: (_verify_xcurve, None),
}


def instantiate_named_group(group_name, conn, recorder):
    """Create a KeyExchange object according to the given group name.

    Arguments:
        group_name (:class:`tlsmate.tls.SupportedGroups`): the group name
        conn (:obj:`tlsmate.connection.TlsConnection`): a connection object
        recorder (:obj:`tlsmate.recorder.Recorder`): the recorder object

    Returns:
        :obj:`KeyExchange`: the created object
    """

    group = _supported_groups.get(group_name)
    if group is None:
        raise CurveNotSupportedError(
            f"the group {group_name} is not supported", group_name
        )

    kwargs = {"group_name": group_name}
    if group.algo is not None:
        kwargs["algo"] = group.algo

    return group.cls(conn, recorder, **kwargs)


class KeyExchange(metaclass=abc.ABCMeta):
    """The abstract class to derive different key exchange classes from
    """

    def __init__(self, conn, recorder, **kwargs):
        self._group_name = kwargs.get("group_name")
        self._conn = conn
        self._recorder = recorder

    @abc.abstractmethod
    def set_remote_key(self, rem_pub_key, **kwargs):
        """Store the remote public key
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_transferable_key(self):
        """Returns the pdu-format for whatever is transferred to the peer
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_shared_secret(self):
        """Do everything needed to provide the premaster secret
        """

        raise NotImplementedError


class RsaKeyExchange(KeyExchange):
    """Implement an RSA based key transport
    """

    def __init__(self, *args, **kwargs):
        self._pms = None
        super().__init__(*args, **kwargs)

    def set_remote_key(self, rem_pub_key, **kwargs):
        """Not applicable for RSA based ley transport
        """

        raise NotImplementedError

    def _create_pms(self):
        """Build the premaster secret

        The client selects a random number, preceded by the TLS version as sent in
        the ClientHello.
        """
        pms = bytearray()
        version = self._conn.client_version_sent
        val = getattr(version, "value", version)
        pms.extend(pdu.pack_uint16(val))
        random = self._recorder.inject(pms_rsa=os.urandom(46))
        pms.extend(random)
        self._pms = pms

    def get_transferable_key(self):
        """Returns an RSA-encrypted key

        Returns:
            bytes: the encrypted key
        """

        if self._pms is None:
            self._create_pms()

        cert = self._conn.msg.server_certificate.chain.certificates[0]
        rem_pub_key = cert.parsed.public_key()
        ciphered_key = rem_pub_key.encrypt(bytes(self._pms), padding.PKCS1v15())
        # injecting the encrypted key to the recorder is required, as the
        # padding scheme PKCS1v15 produces non-deterministic cipher text.
        return self._recorder.inject(rsa_enciphered=ciphered_key)

    def get_shared_secret(self):
        """Get the shared key, i.e., the premaster secret.

        Returns:
            bytes: the premaster secret
        """

        if self._pms is None:
            self._create_pms()

        return self._pms


class DhKeyExchange(KeyExchange):
    """Implement Diffie-Hellman key exchange
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._group_name is not None:
            dh_nbrs = dh_numbers.dh_numbers.get(self._group_name)
            if dh_nbrs is None:
                raise ValueError(
                    f"No numbers defined for DH-group {self._group_name.name}"
                )

            self._gval = dh_nbrs.g_val
            self._pval = int.from_bytes(dh_nbrs.p_val, "big")

        else:
            self._pval = None
            self._gval = None

        self._rem_pub_key = None
        self._priv_key = None
        self._pub_key = None
        self._dh_group = None

    def _create_key_pair(self):
        self._dh_group = dh.DHParameterNumbers(self._pval, self._gval)
        if self._recorder.is_injecting():
            x = self._recorder.inject(x_val=None)
            self._priv_key = load_der_private_key(x, None)
            self._pub_key = self._priv_key.public_key()

        else:
            self._priv_key = self._dh_group.parameters().generate_private_key()
            self._pub_key = self._priv_key.public_key()
            x = self._priv_key.private_bytes(
                Encoding.DER, PrivateFormat.PKCS8, NoEncryption()
            )
            self._recorder.trace(x_val=x)

    def get_shared_secret(self):
        """Get the shared key, i.e., the premaster secret.

        Returns:
            bytes: the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()

        rem_pub_numbers = dh.DHPublicNumbers(self._rem_pub_key, self._dh_group)
        rem_pub_key = rem_pub_numbers.public_key()
        return self._priv_key.exchange(rem_pub_key).lstrip(b"\0")

    def set_remote_key(self, rem_pub_key, g_val=None, p_val=None, group=None):
        """Sets the remote public key

        Arguments:
            rem_pub_key (bytes): the raw remote public key
            g_val (int): the generator value
            p_val (bytes): the p-value as a byte string
            group (:class:`tlsmate.tls.SupportedGroups`): the supported group
        """

        if group is not None:
            dh_nbrs = dh_numbers.dh_numbers.get(group)
            if dh_nbrs is None:
                raise ServerMalfunction(tls.ServerIssue.FFDH_GROUP_UNKNOWN)

            p_val = dh_nbrs.p_val
            g_val = dh_nbrs.g_val

        self._pval = int.from_bytes(p_val, "big")
        self._gval = g_val
        self._rem_pub_key = int.from_bytes(rem_pub_key, "big")

    def get_transferable_key(self):
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            bytes: the key to be sent to the peer.
        """

        if self._pub_key is None:
            self._create_key_pair()

        y_val = self._pub_key.public_numbers().y
        return y_val.to_bytes(int(self._pub_key.key_size / 8), "big")

    def get_key_share(self):
        """TLS1.3 alias for get_transferable_key.
        """

        return self.get_transferable_key()


class EcdhKeyExchange(KeyExchange):
    """Implement an ECDHE key exchange
    """

    def __init__(self, *args, **kwargs):
        self._algo = kwargs.get("algo")
        self._priv_key = None
        self._pub_key = None
        self._rem_pub_key = None
        super().__init__(*args, **kwargs)

    def _create_key_pair(self):
        seed = int.from_bytes(os.urandom(10), "big")
        seed = self._recorder.inject(ec_seed=seed)
        self._priv_key = ec.derive_private_key(seed, self._algo())
        pub_key = self._priv_key.public_key()
        self._pub_key = pub_key.public_bytes(
            Encoding.X962, PublicFormat.UncompressedPoint
        )

    def get_shared_secret(self):
        """Get the shared key, i.e., the premaster secret.

        Returns:
            bytes: the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()

        rem_pub_key = ec.EllipticCurvePublicKey.from_encoded_point(
            self._algo(), bytes(self._rem_pub_key)
        )
        return self._priv_key.exchange(ec.ECDH(), rem_pub_key)

    def set_remote_key(self, rem_pub_key, **kwargs):
        """Sets the remote public key

        Arguments:
            rem_pub_key (bytes): the raw remote public key
            **kwargs: unused
        """

        self._rem_pub_key = rem_pub_key

    def get_transferable_key(self):
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            bytes: the key to be sent to the peer.
        """

        if self._pub_key is None:
            self._create_key_pair()

        return self._pub_key

    def get_key_share(self):
        """TLS1.3 alias for get_transferable_key.
        """

        if self._pub_key is None:
            self._create_key_pair()

        return self._pub_key


class EcdhKeyExchangeCertificate(object):
    """Implement the key exchange for ECDHE.
    """

    def __init__(self, conn, recorder):
        cert = conn.msg.server_certificate.chain.certificates[0]
        rem_pub_key = cert.parsed.public_key()
        seed = recorder.inject(ec_seed=int.from_bytes(os.urandom(10), "big"))
        priv_key = ec.derive_private_key(seed, rem_pub_key.curve)
        pub_key = priv_key.public_key()
        self._pub_key = pub_key.public_bytes(
            Encoding.X962, PublicFormat.UncompressedPoint
        )
        self._shared_secret = priv_key.exchange(ec.ECDH(), rem_pub_key)

    def set_remote_key(self, rem_pub_key, **kwargs):
        """Sets the remote public key

        Arguments:
            rem_pub_key (bytes): the raw remote public key
            **kwargs: unused
        """

        pass

    def get_transferable_key(self):
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            bytes: the key to be sent to the peer.
        """

        return self._pub_key

    def get_shared_secret(self):
        """Get the shared key, i.e., the premaster secret.

        Returns:
            bytes: the premaster secret
        """

        return self._shared_secret


class XKeyExchange(KeyExchange):
    """Implement the key exchange for X25519 and X448.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._priv_key = None
        self._pub_key = None
        self._rem_pub_key = None
        if self._group_name is tls.SupportedGroups.X25519:
            self._private_key_lib = x25519.X25519PrivateKey
            self._public_key_lib = x25519.X25519PublicKey

        elif self._group_name is tls.SupportedGroups.X448:
            self._private_key_lib = x448.X448PrivateKey
            self._public_key_lib = x448.X448PublicKey

        else:
            raise ValueError(f"the group name {self._group_name.name} is not expected")

    def _create_key_pair(self):
        if self._recorder.is_injecting():
            priv_bytes = self._recorder.inject(private_key=None)
            self._priv_key = self._private_key_lib.from_private_bytes(priv_bytes)

        else:
            self._priv_key = self._private_key_lib.generate()
            if self._recorder.is_recording():
                priv_bytes = self._priv_key.private_bytes(
                    encoding=Encoding.Raw,
                    format=PrivateFormat.Raw,
                    encryption_algorithm=NoEncryption(),
                )
                self._recorder.trace(private_key=priv_bytes)

        pub_key = self._priv_key.public_key()
        self._pub_key = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)

    def get_shared_secret(self):
        """Get the shared key, i.e., the premaster secret.

        Returns:
            bytes: the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()

        rem_pub_key = self._public_key_lib.from_public_bytes(bytes(self._rem_pub_key))
        return self._priv_key.exchange(rem_pub_key)

    def set_remote_key(self, rem_pub_key, **kwargs):
        """Sets the remote public key

        Arguments:
            rem_pub_key (bytes): the raw remote public key
            **kwargs: unused
        """

        self._rem_pub_key = rem_pub_key

    def get_transferable_key(self):
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            bytes: the key to be sent to the peer.
        """

        if self._priv_key is None:
            self._create_key_pair()

        return self._pub_key

    def get_key_share(self):
        """TLS1.3 alias for get_transferable_key.
        """

        return self.get_transferable_key()


class _Group(NamedTuple):
    """Structure for a group
    """

    cls: type
    algo: type


_supported_groups = {
    tls.SupportedGroups.SECT163K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT163K1),
    tls.SupportedGroups.SECT163R2: _Group(cls=EcdhKeyExchange, algo=ec.SECT163R2),
    tls.SupportedGroups.SECT233K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT233K1),
    tls.SupportedGroups.SECT233R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT233R1),
    tls.SupportedGroups.SECT283K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT283K1),
    tls.SupportedGroups.SECT283R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT283R1),
    tls.SupportedGroups.SECT409K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT409K1),
    tls.SupportedGroups.SECT409R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT409R1),
    tls.SupportedGroups.SECT571K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT571K1),
    tls.SupportedGroups.SECT571R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT571R1),
    tls.SupportedGroups.SECP192R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP192R1),
    tls.SupportedGroups.SECP224R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP224R1),
    tls.SupportedGroups.SECP256K1: _Group(cls=EcdhKeyExchange, algo=ec.SECP256K1),
    tls.SupportedGroups.SECP256R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP256R1),
    tls.SupportedGroups.SECP384R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP384R1),
    tls.SupportedGroups.SECP521R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP521R1),
    tls.SupportedGroups.BRAINPOOLP256R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP256R1
    ),
    tls.SupportedGroups.BRAINPOOLP384R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP384R1
    ),
    tls.SupportedGroups.BRAINPOOLP512R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP512R1
    ),
    tls.SupportedGroups.X25519: _Group(cls=XKeyExchange, algo=None),
    tls.SupportedGroups.X448: _Group(cls=XKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE2048: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE3072: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE4096: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE6144: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE8192: _Group(cls=DhKeyExchange, algo=None),
}



# -*- coding: utf-8 -*-
"""Module providing classes for each TLS message.
"""
# import basic stuff
import abc
import os

# import own stuff
from tlsmate import tls
from tlsmate import ext
from tlsmate.exception import ServerMalfunction
from tlsmate.cert_chain import CertChain
from tlsmate import pdu

# import other stuff


def get_extension(extensions, ext_id):
    """Helper function to search for an extension

    Arguments:
        extensions (list of :obj:`tlsmate.tls.Extensions`): the list
            to search for the extension
        ext_id (:obj:`tlsmate.tls.Extensions`): the extension to to look for

    Returns:
        :obj:`tlsmate.tls.Extensions`: The extension if present, or None
            otherwise.
    """

    if extensions is not None:
        for extension in extensions:
            if extension.extension_id == ext_id:
                return extension

    return None


def _deserialize_extensions(extensions, fragment, offset):
    """Helper function to deserialize extensions

    Arguments:
        extensions (list): the list where to store the deserialized extensions
        fragment (bytes): the pdu buffer
        offset (int): the offset within the pdu buffer to the start of the extensions

    Returns:
        int: the offset on the byte following the extensions
    """

    if offset < len(fragment):
        # extensions present
        ext_len, offset = pdu.unpack_uint16(fragment, offset)
        ext_end = offset + ext_len
        while offset < ext_end:
            extension, offset = ext.Extension.deserialize(fragment, offset)
            extensions.append(extension)

    return offset


def _get_version(msg):
    """Get the highest version from a hello message.

    Arguments:
        msg (:obj:`HandshakeMessage`): the message

    Returns:
        :class:`tls.Version`: The highest version
    """

    supported_versions = msg.get_extension(tls.Extension.SUPPORTED_VERSIONS)
    if supported_versions is not None:
        return supported_versions.versions[0]

    return msg.version


class TlsMessage(metaclass=abc.ABCMeta):
    """Abstract base class for TLS messages
    """

    @abc.abstractmethod
    def deserialize(cls, fragment, conn):
        """Method to deserialize the message received from the network

        Arguments:
            fragment (bytearray): The byte array representation of the message in
                network order (big endian).
            conn (:obj:`tlsmate.connection.TlsConnection`): The connection object,
                needed for some odd cases, e.g. when deserializing a ServerKeyExchange
                message, as the layout purely depends on the key exchange method.

        Returns:
            :obj:`TlsMessage`:
                The deserialized message represented in a python object.
        """

        pass

    @abc.abstractmethod
    def serialize(self, conn):
        """A method to serialize this object.

        Arguments:
            conn (:obj:`tlsmate.connection.TlsConnection`): The connection object,
                needed for some odd cases, e.g. when serializing a ServerKeyExchange
                message, as the layout purely depends on the key exchange method.

        Returns:
            bytearray: The bytes of the message in network order (big endian).
        """

        pass


class Any(object):
    """Class to represent any message to wait for in a test case.
    """

    pass


class Timeout(object):
    """Class to allow waiting for a timeout.
    """

    msg_type = "Timeout"


class HandshakeMessage(TlsMessage):
    """A base class for all handshake messages.
    """

    content_type = tls.ContentType.HANDSHAKE
    """ :obj:`tlsmate.tls.ContentType.HANDSHAKE`
    """

    msg_type = None
    """ :obj:`tlsmate.tls.HandshakeType`: The type of the handshake message.
    """

    @classmethod
    def deserialize(cls, fragment, conn):
        msg_type, offset = pdu.unpack_uint8(fragment, 0)
        msg_type = tls.HandshakeType.val2enum(msg_type, alert_on_failure=True)
        length, offset = pdu.unpack_uint24(fragment, offset)
        if length + offset != len(fragment):
            raise ServerMalfunction(
                tls.ServerIssue.MESSAGE_LENGTH_ERROR, message=cls.msg_type
            )

        cls_name = _hs_deserialization_map[msg_type]
        msg = cls_name()._deserialize_msg_body(fragment, offset, conn)
        return msg

    @abc.abstractmethod
    def _deserialize_msg_body(self, msg_body, offset, conn):
        """Method to deserialize a handshake message.

        Arguments:
            msg_body (bytearray): The bytes which contain the serialized handshake
                message.
            offset (int): The offset within the bytes where the handshake message
                starts.
            conn (:obj:`tlsmate.connection.TlsConnection`): the connection object,
                used to retrieve additional information required to deserialize
                the handshake message.

        Returns:
            :obj:`HandshakeMessage`: the deserialized message object, i.e. self.
        """

        pass

    @abc.abstractmethod
    def _serialize_msg_body(self, conn):
        """Serializes the message body.

        I.e., the message body only is serialized, without the common handshake header.
        """

        pass

    def serialize(self, conn):
        msg_body = self._serialize_msg_body(conn)

        return bytearray(
            pdu.pack_uint8(self.msg_type.value)
            + pdu.pack_uint24(len(msg_body))
            + msg_body
        )


class HelloRequest(HandshakeMessage):
    """This class represents a HelloRequest message.
    """

    msg_type = tls.HandshakeType.HELLO_REQUEST
    """:obj:`tlsmate.tls.HandshakeType.HELLO_REQUEST`
    """

    def __init__(self):
        pass

    def _serialize_msg_body(self, conn):
        return b""

    def _deserialize_msg_body(self, fragment, offset, conn):
        if offset != len(fragment):
            raise ServerMalfunction(
                tls.ServerIssue.MESSAGE_LENGTH_ERROR, message=self.msg_type
            )

        return self


class ClientHello(HandshakeMessage):
    """This class represents a ClientHello message.

    Attributes:
        version (:obj:`tlsmate.tls.Version`):
            The version the client offers to the server.
        random (bytes): The random value.
        session_id (bytes): The session id.
        cipher_suites (list of :obj:`tlsmate.tls.CipherSuite`): The list
            of cipher suites offered to the server
        compression_methods (list of :obj:`tlsmate.tls.CompressionMethod`):
            The list of compression methods offered to the server.
        extensions (list of :obj:`tlsmate.ext.Extension`): The list of
            extensions offered to the server. It can be a class (in this case the
            content of the extension is filled according to the client profile), or
            an instance, which leaves the user the full control of the contents.
    """

    msg_type = tls.HandshakeType.CLIENT_HELLO
    """:obj:`tlsmate.tls.HandshakeType.CLIENT_HELLO`
    """

    def __init__(self):
        self.version = tls.Version.TLS12
        self.random = None
        self.session_id = bytes()
        self.cipher_suites = [tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256]
        self.compression_methods = [tls.CompressionMethod.NULL]
        self.extensions = []
        self._bytes_after_psk_ext = 0

    def _serialize_msg_body(self, conn):
        msg = bytearray()

        # version
        msg.extend(pdu.pack_uint16(getattr(self.version, "value", self.version)))

        # random
        msg.extend(self.random)

        # session_id
        msg.extend(pdu.pack_uint8(len(self.session_id)))
        msg.extend(self.session_id)

        # cipher suites
        msg.extend(pdu.pack_uint16(2 * len(self.cipher_suites)))
        for cipher_suite in self.cipher_suites:
            if type(cipher_suite) == int:
                msg.extend(pdu.pack_uint16(cipher_suite))

            else:
                msg.extend(pdu.pack_uint16(cipher_suite.value))

        # compression methods
        msg.extend(pdu.pack_uint8(len(self.compression_methods)))
        for comp_meth in self.compression_methods:
            if type(comp_meth) == tls.CompressionMethod:
                comp_meth = comp_meth.value

            msg.extend(pdu.pack_uint8(comp_meth))

        # extensions
        if self.extensions is not None:
            psk_end_offset = 0
            ext_bytes = bytearray()
            for extension in self.extensions:
                ext_bytes.extend(extension.serialize(conn))
                if extension.extension_id is tls.Extension.PRE_SHARED_KEY:
                    psk_end_offset = len(ext_bytes)

            msg.extend(pdu.pack_uint16(len(ext_bytes)))
            msg.extend(ext_bytes)
            self._bytes_after_psk_ext = len(ext_bytes) - psk_end_offset

        return msg

    def _deserialize_msg_body(self, msg_body, length, conn):
        return self

    def get_extension(self, ext_id):
        """Method to extract a specific extension.

        Arguments:
            ext_id (:obj:`tlsmate.tls.Extension`): The extension id to
                look for.

        Returns:
            :obj:`tlsmate.ext.Extension`:
                The extension object or None if not present.
        """

        return get_extension(self.extensions, ext_id)

    def get_version(self):
        """Get the highest TLS version from the message.

        Takes the version parameter into account as well as the extension
        SUPPORTED_VERSIONS (if present).

        Returns:
            :class:`tlsmate.tls.Version`: The highest TLS version offered.
        """

        return _get_version(self)


class ServerHello(HandshakeMessage):
    """This class represents a ServerHello message.

    Attributes:
        version (:obj:`tlsmate.tls.Version`): The version selected by
            the server.
        random (bytes): The random value from the server.
        session_id (bytes): The session id from the server.
        compression_method (:obj:`tlsmate.tls.CompressionMethod`): The
            selected compression method by the server.
        extensions (list of :obj:`tlsmate.ext.Extension`): The list of
            extensions as returned from the server.
    """

    msg_type = tls.HandshakeType.SERVER_HELLO
    """:obj:`tlsmate.tls.HandshakeType.SERVER_HELLO`
    """

    HELLO_RETRY_REQ_RAND = bytes.fromhex(
        "CF 21 AD 74 E5 9A 61 11 BE 1D 8C 02 1E 65 B8 91 "
        "C2 A2 11 16 7A BB 8C 5E 07 9E 09 E2 C8 A8 33 9C "
    )

    def __init__(self):
        self.version = None
        self.random = None
        self.session_id = None
        self.cipher_suite = None
        self.compression_method = None
        self.extensions = []

    def _serialize_msg_body(self, conn):
        # TODO if we want to implement the server side as well
        pass

    def _deserialize_msg_body(self, fragment, offset, conn):
        version, offset = pdu.unpack_uint16(fragment, offset)
        self.version = tls.Version.val2enum(version, alert_on_failure=True)
        self.random, offset = pdu.unpack_bytes(fragment, offset, 32)
        if self.random == self.HELLO_RETRY_REQ_RAND:
            self.msg_type = tls.HandshakeType.HELLO_RETRY_REQUEST

        session_id_len, offset = pdu.unpack_uint8(fragment, offset)
        self.session_id, offset = pdu.unpack_bytes(fragment, offset, session_id_len)
        cipher_suite, offset = pdu.unpack_uint16(fragment, offset)
        self.cipher_suite = tls.CipherSuite.val2enum(
            cipher_suite, alert_on_failure=True
        )
        compression_method, offset = pdu.unpack_uint8(fragment, offset)
        self.compression_method = tls.CompressionMethod.val2enum(
            compression_method, alert_on_failure=True
        )
        self.extensions = []
        _deserialize_extensions(self.extensions, fragment, offset)
        return self

    def get_extension(self, ext_id):
        """Get an extension from the message

        Arguments:
            ext_id: (:class:`tlsmate.tls`): The extensions to look for

        Returns:
            :obj:`tlsmate.ext.Extension`: The extension or None if not present.
        """

        return get_extension(self.extensions, ext_id)

    def get_version(self):
        """Get the negotiated TLS version from the message.

        Takes the version parameter into account as well as the extension
        SUPPORTED_VERSIONS (if present).

        Returns:
            :class:`tlsmate.tls.Version`: The negotiated TLS version.
        """

        return _get_version(self)


ServerHello.get_extension.__doc__ = ClientHello.get_extension.__doc__


class Certificate(HandshakeMessage):
    """This class represents a Certificate message.

    The TLS1.3 format is not yet 100% implemented, but the basic certificate
    chain (as for TLS1.2 and below) is supported.

    Attributes:
        request_context (bytes): Set to None for TLS1.2 and below. For TLS1.3
            it is only applicable for Certificate messages send by the client. TLS1.3
            servers SHALL set this to a zero-length field (RFC8446, 4.4.2)
        certificates (list of bytes): The certificate chain, as received from the
            peer, i.e. the list will start with the host certificate.
    """

    msg_type = tls.HandshakeType.CERTIFICATE
    """:obj:`tlsmate.tls.HandshakeType.CERTIFICATE`
    """

    def __init__(self):
        self.request_context = None
        self.chain = CertChain()
        self.extensions = []

    def _serialize_msg_body(self, conn):
        msg = bytearray()
        if self.request_context is not None:
            msg.extend(pdu.pack_uint8(len(self.request_context)))
            msg.extend(self.request_context)

        cert_list = bytearray()
        if self.chain is not None:
            for certificate in self.chain.certificates:
                cert_list.extend(pdu.pack_uint24(len(certificate.bytes)))
                cert_list.extend(certificate.bytes)
                if conn.version is tls.Version.TLS13:
                    # no extensions supported right now, set length to 0.
                    cert_list.extend(pdu.pack_uint16(0))

        msg.extend(pdu.pack_uint24(len(cert_list)))
        msg.extend(cert_list)
        return msg

    def _deserialize_msg_body(self, fragment, offset, conn):

        if conn.version is tls.Version.TLS13:
            length, offset = pdu.unpack_uint8(fragment, offset)
            if length:
                self.request_context, offset = pdu.unpack_bytes(
                    fragment, offset, length
                )

        list_len, offset = pdu.unpack_uint24(fragment, offset)
        while offset < len(fragment):
            cert_len, offset = pdu.unpack_uint24(fragment, offset)
            certificate, offset = pdu.unpack_bytes(fragment, offset, cert_len)
            self.chain.append_bin_cert(certificate)
            if conn.version is tls.Version.TLS13:
                offset = _deserialize_extensions(
                    self.chain.certificates[-1].tls_extensions, fragment, offset
                )

        return self


class CertificateVerify(HandshakeMessage):
    """This class represents a CertificateVerify message.

    Attributes:
        signature_scheme (:obj:`tlsmate.tls.SignatureScheme`): The scheme
            used for the signature
        signature (bytes): The signature.
    """

    msg_type = tls.HandshakeType.CERTIFICATE_VERIFY
    """:obj:`tlsmate.tls.HandshakeType.CERTIFICATE_VERIFY`
    """

    def __init__(self):
        self.signature_scheme = None
        self.signature = None

    def _serialize_msg_body(self, conn):
        msg = bytearray()
        msg.extend(pdu.pack_uint16(self.signature_scheme.value))
        msg.extend(pdu.pack_uint16(len(self.signature)))
        msg.extend(self.signature)
        return msg

    def _deserialize_msg_body(self, fragment, offset, conn):
        scheme, offset = pdu.unpack_uint16(fragment, offset)
        self.signature_scheme = tls.SignatureScheme(scheme)
        length, offset = pdu.unpack_uint16(fragment, offset)
        self.signature, offset = pdu.unpack_bytes(fragment, offset, length)
        return self


class KeyExchangeEC(object):
    """Class representing parameters for a DH-based key exchange.

    Currently only named curves are supported.

    Attributes:
        curve_type (:obj:`tlsmate.tls.EcCurveType`): The type of the curve.
        named_curve (:obj:`tlsmate.tls.SupportedGroups`): The curve name, as
            defined by the supported groups.
        public (bytes): The public key from the peer.
        sig_scheme (:obj:`tlsmate.tls.SignatureScheme`): The scheme of the
            signature, if present.
        signature (bytes): The signature.
        signed_params (bytes): The part of the message which has been signed.
    """

    def __init__(self):
        self.curve_type = None
        self.named_curve = None
        self.public = None
        self.sig_scheme = None
        self.signature = None
        self.signed_params = None

    def _deserialize_ECParameters(self, fragment, offset, conn):
        """Deserializes the EC-parameters.

        Arguments:
            fragment (bytes): A PDU buffer as received from the network.
            offset (int): The offset within the fragment where the DH-params start.
            conn (:obj:`tlsmate.connection.TlsConnection`): The connection object.
        Returns:
            int: The new offset after unpacking the EC-parameters.
        """

        curve_type, offset = pdu.unpack_uint8(fragment, offset)
        self.curve_type = tls.EcCurveType.val2enum(curve_type)
        if self.curve_type is tls.EcCurveType.NAMED_CURVE:
            named_curve, offset = pdu.unpack_uint16(fragment, offset)
            self.named_curve = tls.SupportedGroups.val2enum(named_curve)

        # TODO: add other curve types
        return offset

    def _deserialize_ServerECDHParams(self, fragment, offset, conn):
        """Deserializes the ECDH-parameters.

        Arguments:
            fragment (bytes): A PDU buffer as received from the network.
            offset (int): The offset within the fragment where the DH-params start.
            conn (:obj:`tlsmate.connection.TlsConnection`): The connection object.
        Returns:
            int: The new offset after unpacking the EC-parameters.
        """

        # ECParameters    curve_params;
        offset = self._deserialize_ECParameters(fragment, offset, conn)
        # ECPoint         public;
        point_len, offset = pdu.unpack_uint8(fragment, offset)
        self.public, offset = pdu.unpack_bytes(fragment, offset, point_len)

        return offset

    def _deserialize_msg_body(self, fragment, offset, conn):
        """Deserializes the EC-parameters.

        Arguments:
            fragment (bytes): A PDU buffer as received from the network.
            offset (int): The offset within the fragment where the DH-params start.
            conn (:obj:`tlsmate.connection.TlsConnection`): The connection object.
        Returns:
            :obj:`KeyExchangeEC`: The object representing the EC parameters, i.e. self.
        """

        signed_params_start = offset
        # ServerECDHParams    params;
        offset = self._deserialize_ServerECDHParams(fragment, offset, conn)
        self.signed_params = fragment[signed_params_start:offset]
        # Signature           signed_params;
        if conn.cs_details.key_algo is not tls.KeyExchangeAlgorithm.ECDH_ANON:
            if conn.version is tls.Version.TLS12:
                sig_scheme, offset = pdu.unpack_uint16(fragment, offset)
                self.sig_scheme = tls.SignatureScheme.val2enum(sig_scheme)

            signature_len, offset = pdu.unpack_uint16(fragment, offset)
            self.signature, offset = pdu.unpack_bytes(fragment, offset, signature_len)

        return self


class KeyExchangeDH(object):
    """Class representing parameters for a DH-based key exchange.

    Attributes:
        p_val (bytes): The modulo p used for the DH-key exchange.
        g_val (int): The generator g used for the DH.key exchange.
        public_key (bytes): The public key of the peer.
        sig_scheme (:obj:`tlsmate.tls.SignatureScheme`): the signature
            scheme use in the signature. Only used for authenticated DH, set to None
            for anonymous key exchange.
        signature (bytes): The signature. Only used for authenticated DH, set to None
            for anonymous key exchange.
    """

    def __init__(self):
        self.p_val = None
        self.g_val = None
        self.public_key = None
        self.sig_scheme = None
        self.signature = None
        self.signed_params = None

    def _deserialize_msg_body(self, fragment, offset, conn, signature_present=True):
        signed_params_start = offset
        p_length, offset = pdu.unpack_uint16(fragment, offset)
        self.p_val, offset = pdu.unpack_bytes(fragment, offset, p_length)
        g_len, offset = pdu.unpack_uint16(fragment, offset)
        g_bytes, offset = pdu.unpack_bytes(fragment, offset, g_len)
        self.g_val = int.from_bytes(g_bytes, "big")
        pub_key_len, offset = pdu.unpack_uint16(fragment, offset)
        self.public_key, offset = pdu.unpack_bytes(fragment, offset, pub_key_len)
        if signature_present:
            self.signed_params = fragment[signed_params_start:offset]
            if conn.version is tls.Version.TLS12:
                sig_scheme, offset = pdu.unpack_uint16(fragment, offset)
                self.sig_scheme = tls.SignatureScheme.val2enum(sig_scheme)

            sig_length, offset = pdu.unpack_uint16(fragment, offset)
            self.signature, offset = pdu.unpack_bytes(fragment, offset, sig_length)

        return self


class ServerKeyExchange(HandshakeMessage):
    """This class represents a ServerKeyExchange message.

    The available attributes depend on the key exchange type. If they are not
    applicable, they are set to None.

    Attributes:
        ec (:obj:`KeyExchangeEC`): Object representing the parameters for an EC-based
            key exchange.
        dh (:obj:`KeyExchangeDH`): Object representing the parameters for an DH-based
            key exchange.
    """

    msg_type = tls.HandshakeType.SERVER_KEY_EXCHANGE
    """:obj:`tlsmate.tls.HandshakeType.SERVER_KEY_EXCHANGE`
    """

    def __init__(self):
        self.ec = None
        self.dh = None

    def _serialize_msg_body(self, conn):
        # TODO if we want to implement the server side as well
        pass

    def _deserialize_msg_body(self, fragment, offset, conn):
        self.ec = None
        self.dh = None
        if conn.cs_details.key_algo_struct.key_ex_type is tls.KeyExchangeType.ECDH:
            # RFC 4492
            self.ec = KeyExchangeEC()._deserialize_msg_body(fragment, offset, conn)

        elif conn.cs_details.key_algo_struct.key_ex_type is tls.KeyExchangeType.DH:
            # RFC5246
            self.dh = KeyExchangeDH()._deserialize_msg_body(
                fragment,
                offset,
                conn,
                signature_present=(
                    conn.cs_details.key_algo_struct.key_auth
                    is not tls.KeyAuthentication.NONE
                ),
            )

        else:
            raise ServerMalfunction(tls.ServerIssue.INCOMPATIBLE_KEY_EXCHANGE)

        return self


class ServerHelloDone(HandshakeMessage):
    """This class represents a ServerHelloDone message.
    """

    msg_type = tls.HandshakeType.SERVER_HELLO_DONE
    """:obj:`tlsmate.tls.HandshakeType.SERVER_HELLO_DONE`
    """

    def __init__(self):
        pass

    def _serialize_msg_body(self, conn):
        # TODO if we want to implement the server side as well
        pass

    def _deserialize_msg_body(self, fragment, offset, conn):
        if offset != len(fragment):
            raise ServerMalfunction(
                tls.ServerIssue.MESSAGE_LENGTH_ERROR, message=self.msg_type
            )

        return self


class ClientKeyExchange(HandshakeMessage):
    """This class represents a ClientKeyExchange message.

    The attributes depend on the key exchange type.

    Attributes:
        rsa_encrypted_pms (bytes): The premaster secret encrypted with RSA. Only
            applicable for RSA-based key transport.
        dh_public (bytes): The public DH key of the client. Only applicable for
            DH key exchanges.
        ecdh_public (bytes): The public ECDH key of the client. Only applicable for
            for ECDH key exchanges.
    """

    msg_type = tls.HandshakeType.CLIENT_KEY_EXCHANGE
    """:obj:`tlsmate.tls.HandshakeType.CLIENT_KEY_EXCHANGE`
    """

    def __init__(self):
        self.rsa_encrypted_pms = None
        self.dh_public = None
        self.ecdh_public = None

    def _serialize_msg_body(self, conn):
        msg = bytearray()
        if self.rsa_encrypted_pms is not None:
            if conn.version is not tls.Version.SSL30:
                msg.extend(pdu.pack_uint16(len(self.rsa_encrypted_pms)))

            msg.extend(self.rsa_encrypted_pms)

        elif self.dh_public is not None:
            msg.extend(pdu.pack_uint16(len(self.dh_public)))
            msg.extend(self.dh_public)

        elif self.ecdh_public is not None:
            msg.extend(pdu.pack_uint8(len(self.ecdh_public)))
            msg.extend(self.ecdh_public)

        return msg

    def _deserialize_msg_body(self, fragment, offset, conn):
        pass


class Finished(HandshakeMessage):
    """This class represents a Finished message.

    Attributes:
        verify_data (bytes): The verify data item.
    """

    msg_type = tls.HandshakeType.FINISHED
    """:obj:`tlsmate.tls.HandshakeType.FINISHED`
    """

    def __init__(self):
        self.verify_data = None

    def _serialize_msg_body(self, conn):
        return bytes(self.verify_data)

    def _deserialize_msg_body(self, fragment, offset, conn):
        self.verify_data = fragment[offset:]
        return self


class EndOfEarlyData(HandshakeMessage):
    """This class represents an EndOfEarlyData message.
    """

    msg_type = tls.HandshakeType.END_OF_EARLY_DATA
    """:obj:`tlsmate.tls.HandshakeType.END_OF_EARLY_DATA`
    """

    def __init__(self):
        pass

    def _serialize_msg_body(self, conn):
        return b""

    def _deserialize_msg_body(self, fragment, offset, conn):
        if offset != len(fragment):
            raise ServerMalfunction(
                tls.ServerIssue.MESSAGE_LENGTH_ERROR, message=self.msg_type
            )

        return self


class NewSessionTicket(HandshakeMessage):
    """This class represents a NewSessionTicket message.

    Attributes:
        lifetime (int): The lifetime hint for the ticket in seconds.
        age_add (int): The age_add parameter (only TLS1.3)
        nonce (bytes): The nonce (only TLS1.3)
        ticket (bytes): The ticket.
        extensions (list of :obj:`tlsmate.tls.Extension`): The list of
            TLS extensions (only TLS1.3).
    """

    msg_type = tls.HandshakeType.NEW_SESSION_TICKET
    """:obj:`tlsmate.tls.HandshakeType.NEW_SESSION_TICKET`
    """

    def __init__(self):
        self.lifetime = None
        self.age_add = None
        self.nonce = None
        self.ticket = None
        self.extensions = []

    def _serialize_msg_body(self, conn):
        # TODO for server side implementation
        return bytearray()

    def _deserialize_msg_body(self, fragment, offset, conn):
        if conn.version is tls.Version.TLS13:
            self.lifetime, offset = pdu.unpack_uint32(fragment, offset)
            self.age_add, offset = pdu.unpack_uint32(fragment, offset)
            length, offset = pdu.unpack_uint8(fragment, offset)
            self.nonce, offset = pdu.unpack_bytes(fragment, offset, length)
            length, offset = pdu.unpack_uint16(fragment, offset)
            self.ticket, offset = pdu.unpack_bytes(fragment, offset, length)
            _deserialize_extensions(self.extensions, fragment, offset)
            return self

        else:
            self.lifetime, offset = pdu.unpack_uint32(fragment, offset)
            length, offset = pdu.unpack_uint16(fragment, offset)
            self.ticket, offset = pdu.unpack_bytes(fragment, offset, length)
            return self

    def get_extension(self, ext_id):
        return get_extension(self.extensions, ext_id)


NewSessionTicket.get_extension.__doc__ = ClientHello.get_extension.__doc__


class CertificateRequest(HandshakeMessage):
    """This class represents a CertificateRequest message.

    Attributes:
        certificate_types (list of :obj:`tlsmate.tls.CertType`): the list of certificate
            types the client may offer. Only used for TLS1.2 and below.
        supported_signature_algorithms (list of :obj:`tlsmate.tls.SignatureScheme`):
            the list of signature algorithms supported by the server. Only used
            for TLS1.2 and below.
        certificate_authorities (list of bytes): the list of acceptable authorities.
            Currently, only byte strings are provided. Only used for TLS1.2 and below.
        certificate_request_context (bytes): an opaque string (only TLS1.3).
        extensions (list of :obj:`tlsmate.tls.Extension`): The list of
            TLS extensions (only TLS1.3).
    """

    msg_type = tls.HandshakeType.CERTIFICATE_REQUEST
    """:obj:`tlsmate.tls.HandshakeType.CERTIFICATE_REQUEST`
    """

    def __init__(self):
        pass

    def _serialize_msg_body(self, conn):
        # TODO for server side implementation
        return bytearray()

    def _deserialize_msg_body(self, fragment, offset, conn):
        if conn.version is tls.Version.TLS13:
            self.certificate_request_context = None
            self.extensions = []
            length, offset = pdu.unpack_uint8(fragment, offset)
            context, offset = pdu.unpack_bytes(fragment, offset, length)
            self.certificate_request_context = context
            _deserialize_extensions(self.extensions, fragment, offset)

        else:
            self.certificate_types = []
            self.supported_signature_algorithms = []
            self.certificate_authorities = []

            length, offset = pdu.unpack_uint8(fragment, offset)
            end = offset + length
            while offset < end:
                cert_type, offset = pdu.unpack_uint8(fragment, offset)
                self.certificate_types.append(tls.CertType.val2enum(cert_type))

            length, offset = pdu.unpack_uint16(fragment, offset)
            end = offset + length
            while offset < end:
                algo, offset = pdu.unpack_uint16(fragment, offset)
                self.supported_signature_algorithms.append(
                    tls.SignatureScheme.val2enum(algo)
                )

            if offset < len(fragment):
                length, offset = pdu.unpack_uint16(fragment, offset)
                end = offset + length
                while offset < end:
                    # TODO: here we only unpack the whole ASN.1 structure as a
                    # byte string.
                    # Unpacking the ASN.1 structure is required.
                    length, offset = pdu.unpack_uint16(fragment, offset)
                    name, offset = pdu.unpack_bytes(fragment, offset, length)
                    self.certificate_authorities.append(name)

        return self

    def get_extension(self, ext_id):
        if not hasattr(self, "extensions"):
            return None

        return get_extension(self.extensions, ext_id)


CertificateRequest.get_extension.__doc__ = ClientHello.get_extension.__doc__


class EncryptedExtensions(HandshakeMessage):
    """This class represents an EncryptedExtensions message.

    Attributes:
        extensions (list of :obj:`tlsmate.ext.Extension`): The list of
            extensions.
    """

    msg_type = tls.HandshakeType.ENCRYPTED_EXTENSIONS

    def __init__(self):
        self.extensions = []

    def _serialize_msg_body(self, conn):
        # TODO for server side implementation
        return bytearray()

    def _deserialize_msg_body(self, fragment, offset, conn):
        _deserialize_extensions(self.extensions, fragment, offset)
        return self

    def get_extension(self, ext_id):
        return get_extension(self.extensions, ext_id)


EncryptedExtensions.get_extension.__doc__ = ClientHello.get_extension.__doc__


class CertificateStatus(HandshakeMessage):
    """This class represents a Certificate Status message.
    """

    msg_type = tls.HandshakeType.CERTIFICATE_STATUS

    def __init__(self,):
        self.status_type = tls.StatusType.OCSP
        self.responses = []

    def _serialize_msg_body(self, conn):
        # TODO for server side implementation
        return bytearray()

    def _unpack_ocsp_response(self, fragment, offset):
        length, offset = pdu.unpack_uint24(fragment, offset)
        response, offset = pdu.unpack_bytes(fragment, offset, length)
        self.responses.append(response)
        return offset

    def _deserialize_msg_body(self, fragment, offset, conn):
        status_type, offset = pdu.unpack_uint8(fragment, offset)
        self.status_type = tls.StatusType.val2enum(status_type)
        if self.status_type is tls.StatusType.OCSP:
            offset = self._unpack_ocsp_response(fragment, offset)

        else:
            length, offset = pdu.unpack_uint24(fragment, offset)
            end = offset + length
            while offset < end:
                offset = self._unpack_ocsp_response(fragment, offset)

        return self


class ChangeCipherSpecMessage(TlsMessage):
    """A base class for all ChangeCipherSpec messages.

    Well, there is only one message for this protocol: ChangeCipherSpec. :-)
    """

    content_type = tls.ContentType.CHANGE_CIPHER_SPEC
    """ :obj:`tlsmate.tls.ContentType.CHANGE_CIPHER_SPEC`
    """

    msg_type = None
    """ :obj:`tlsmate.tls.CCSType`: The type of the CCS message.
    """

    @classmethod
    def deserialize(cls, fragment, conn):
        if len(fragment) != 1:
            raise ServerMalfunction(
                tls.ServerIssue.MESSAGE_LENGTH_ERROR, message=cls.msg_type
            )

        msg_type, offset = pdu.unpack_uint8(fragment, 0)
        msg_type = tls.CCSType.val2enum(msg_type, alert_on_failure=True)
        cls_name = _ccs_deserialization_map[msg_type]
        msg = cls_name()
        msg._deserialize_msg_body(conn)
        return msg

    def serialize(self, conn):
        self._serialize_msg_body(conn)
        ccs_msg = bytearray()
        ccs_msg.extend(pdu.pack_uint8(self.msg_type.value))
        return ccs_msg

    @abc.abstractmethod
    def _deserialize_msg_body(self, conn):
        """Method to deserialize a CCS message.

        Arguments:
            msg_body (bytearray): The bytes which contain the serialized handshake
                message.
            offset (int): The offset within the bytes where the handshake message
                starts.
            conn (:obj:`tlsmate.connection.TlsConnection`): the connection object,
                used to retrieve additional information required to deserialize
                the handshake message.

        Returns:
            :obj:`ChangeCipherSpecMessage`: the deserialized message object.
        """

        pass

    @abc.abstractmethod
    def _serialize_msg_body(self, conn):
        """Serializes the message body.

        I.e., the message body only is serialized, without the common handshake header.
        """

        pass


class ChangeCipherSpec(ChangeCipherSpecMessage):
    """This class represents a ChangeCipherSpecMessage message.
    """

    msg_type = tls.CCSType.CHANGE_CIPHER_SPEC
    """:obj:`tlsmate.tls.CCSType.CHANGE_CIPHER_SPEC`
    """

    def __init__(self):
        pass

    def _deserialize_msg_body(self, conn):
        pass

    def _serialize_msg_body(self, conn):
        pass


class Alert(TlsMessage):
    """This class represents an Alert message.

    Attributes:
        level (:obj:`tlsmate.tls.AlertLevel`): The level of the alert.
        description(:obj:`tlsmate.tls.AlertDescription`): The description of the alert.
    """

    content_type = tls.ContentType.ALERT
    msg_type = tls.ContentType.ALERT

    def __init__(self, **kwargs):
        self.level = kwargs.get("level", tls.AlertLevel.FATAL)
        self.description = kwargs.get(
            "description", tls.AlertDescription.HANDSHAKE_FAILURE
        )

    def serialize(self, conn):
        alert = bytearray()
        if type(self.level) is int:
            alert.extend(pdu.pack_uint8(self.level))

        else:
            alert.extend(pdu.pack_uint8(self.level.value))

        if type(self.description) is int:
            alert.extend(pdu.pack_uint8(self.description))

        else:
            alert.extend(pdu.pack_uint8(self.description.value))

        return bytes(alert)

    @classmethod
    def deserialize(cls, fragment, conn):
        msg = cls()
        alert_level, offset = pdu.unpack_uint8(fragment, 0)
        msg.level = tls.AlertLevel.val2enum(alert_level)
        descr, offset = pdu.unpack_uint8(fragment, offset)
        msg.description = tls.AlertDescription(descr)
        return msg


class AppDataMessage(TlsMessage):
    """Base class for all AppDataMessage.
    """

    content_type = tls.ContentType.APPLICATION_DATA
    """ :obj:`tlsmate.tls.ContentType.APPLICATION_DATA`
    """

    msg_type = tls.ContentType.APPLICATION_DATA
    """ :obj:`tlsmate.tls.ContentType.APPLICATION_DATA`
    """

    @classmethod
    def deserialize(cls, fragment, conn):
        msg = AppData()
        msg._deserialize_msg_body(fragment, conn)
        return msg

    def serialize(self, conn):
        return self._serialize_msg_body(conn)

    @abc.abstractmethod
    def _deserialize_msg_body(self, conn):
        pass

    @abc.abstractmethod
    def _serialize_msg_body(self, conn):
        pass


class AppData(AppDataMessage):
    """This class represents an AppData message.

    Attributes:
        data (bytes): The application data received/to be sent.
    """

    def __init__(self, *content):
        self.data = bytearray()
        for data in content:
            self.data.extend(data)

    def _deserialize_msg_body(self, fragment, conn):
        self.data = fragment

    def _serialize_msg_body(self, conn):
        return bytes(self.data)


class HeartbeatMessage(TlsMessage):
    """A base class for all Heartbeat messages.
    """

    content_type = tls.ContentType.HEARTBEAT
    """ :obj:`tlsmate.tls.ContentType.HEARTBEAT`
    """

    msg_type = None
    """ :obj:`tlsmate.tls.HeartbeatType`: The type of the Heartbeat message.
    """

    def __init__(self, payload_length=None, payload=None, padding=None):
        self.payload_length = payload_length
        self.payload = payload
        self.padding = padding

    @classmethod
    def deserialize(cls, fragment, conn):
        msg_type, offset = pdu.unpack_uint8(fragment, 0)
        msg_type = tls.HeartbeatType.val2enum(msg_type, alert_on_failure=True)
        payload_length, offset = pdu.unpack_uint16(fragment, offset)
        payload, offset = pdu.unpack_bytes(fragment, offset, payload_length)
        padding = fragment[offset:]
        cls_name = _heartbeat_deserialization_map[msg_type]
        return cls_name(payload_length, payload, padding)

    def serialize(self, conn):
        message = bytearray()
        message.extend(pdu.pack_uint8(self.msg_type.value))
        message.extend(pdu.pack_uint16(self.payload_length))
        message.extend(self.payload)
        message.extend(self.padding)
        return message


class HeartbeatRequest(HeartbeatMessage):
    """This class represents a heartbeat request message.
    """

    msg_type = tls.HeartbeatType.HEARTBEAT_REQUEST
    """ :obj:`tlsmate.tls.HeartbeatType.HEARTBEAT_REQUEST`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HeartbeatResponse(HeartbeatMessage):
    """This class represents a heartbeat response message.
    """

    msg_type = tls.HeartbeatType.HEARTBEAT_RESPONSE
    """ :obj:`tlsmate.tls.HeartbeatType.HEARTBEAT_RESPONSE`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SSL2Message(TlsMessage):
    """A base class for all SSL2 messages.
    """

    content_type = tls.ContentType.SSL2
    """:obj:`tlsmate.tls.ContentType.SSL2`
    """

    msg_type = None

    @classmethod
    def deserialize(cls, fragment, conn, length=0):
        msg_type, offset = pdu.unpack_uint8(fragment, 0)
        msg_type = tls.SSLMessagType.val2enum(msg_type, alert_on_failure=True)

        cls_name = _ssl2_deserialization_map[msg_type]
        msg = cls_name()._deserialize_msg_body(fragment, offset, conn)
        return msg

    @abc.abstractmethod
    def _deserialize_msg_body(self, msg_body, length, conn):
        pass

    @abc.abstractmethod
    def serialize(self, conn):
        pass


class SSL2ClientHello(SSL2Message):
    """This class represents an SSL2 CLientHello message.

    Attributes:
        version (:obj:`tlsmate.tls.SSLVersion`): The version "SSL2".
        cipher_specs (list of :obj:`tlsmate.tls.SSLCipherKind`): The list
            of cipher kinds offered by the client.
        session_id (bytes): The session id.
        challenge (bytes): The random bytes by the client.
    """

    msg_type = tls.SSLMessagType.SSL2_CLIENT_HELLO
    """:obj:`tlsmate.tls.SSLMessagType.SSL2_CLIENT_HELLO`
    """

    def __init__(self):
        self.version = tls.SSLVersion.SSL2
        self.cipher_specs = [
            tls.SSLCipherKind.SSL_CK_RC4_128_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_RC4_128_EXPORT40_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_RC2_128_CBC_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_RC2_128_CBC_EXPORT40_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_IDEA_128_CBC_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_DES_64_CBC_WITH_MD5,
            tls.SSLCipherKind.SSL_CK_DES_192_EDE3_CBC_WITH_MD5,
        ]
        self.session_id = bytes()
        self.challenge = os.urandom(16)

    def serialize(self, conn):
        msg = bytearray()
        msg.extend(pdu.pack_uint8(self.msg_type.value))
        msg.extend(pdu.pack_uint16(self.version.value))
        msg.extend(pdu.pack_uint16(3 * len(self.cipher_specs)))
        msg.extend(pdu.pack_uint16(len(self.session_id)))
        msg.extend(pdu.pack_uint16(len(self.challenge)))
        for ck in self.cipher_specs:
            msg.extend(pdu.pack_uint24(ck.value))

        msg.extend(self.session_id)
        msg.extend(self.challenge)
        return msg

    def _deserialize_msg_body(self, fragment, length, conn):
        return self


class SSL2ServerHello(SSL2Message):
    """This class represents an SSL2 ServerHello message.

    Attributes:
        session_id_hit (int): An indication if the session id offered by the client
            will be used.
        cert_type (int): The type of the certificate.
        version (:obj:`tlsmate.tls.SSLVersion`): The version "SSL2".
        cipher_specs (list of :obj:`tlsmate.tls.SSLCipherKind`): The list
            of cipher kinds offered by the server.
        connection_id (bytes): The connection id.
        certificate (bytes): The certificate provided by the server.
    """

    msg_type = tls.SSLMessagType.SSL2_SERVER_HELLO
    """:obj:`tlsmate.tls.SSLMessagType.SSL2_SERVER_HELLO`
    """

    def __init__(self):
        self.session_id_hit = None
        self.cert_type = None
        self.version = None
        self.cipher_specs = []
        self.connection_id = None
        self.certificate = None

    def serialize(self, conn):
        pass

    def _deserialize_msg_body(self, fragment, offset, conn):
        self.session_id_hit, offset = pdu.unpack_uint8(fragment, offset)
        self.cert_type, offset = pdu.unpack_uint8(fragment, offset)
        version, offset = pdu.unpack_uint16(fragment, offset)
        self.version = tls.SSLVersion.val2enum(version)
        cert_len, offset = pdu.unpack_uint16(fragment, offset)
        cipher_len, offset = pdu.unpack_uint16(fragment, offset)
        conn_id_len, offset = pdu.unpack_uint16(fragment, offset)
        self.certificate, offset = pdu.unpack_bytes(fragment, offset, cert_len)
        for _ in range(int(cipher_len / 3)):
            cipher, offset = pdu.unpack_uint24(fragment, offset)
            self.cipher_specs.append(tls.SSLCipherKind.val2enum(cipher))

        self.connection_id, offset = pdu.unpack_bytes(fragment, offset, conn_id_len)
        return self


class SSL2Error(SSL2Message):
    """This class represents an SSL2 Error message.

    Attributes:
        session_id_hit (int): An indication if the session id offered by the client
            will be used.
        cert_type (int): The type of the certificate.
        version (:obj:`tlsmate.tls.SSLVersion`): The version "SSL2".
        cipher_specs (list of :obj:`tlsmate.tls.SSLCipherKind`): The list
            of cipher kinds offered by the server.
        connection_id (bytes): The connection id.
        certificate (bytes): The certificate provided by the server.
    """

    msg_type = tls.SSLMessagType.SSL2_ERROR
    """:obj:`tlsmate.tls.SSLMessagType.SSL2_ERROR`
    """

    def __init__(self):
        self.error = None

    def serialize(self, conn):
        pass

    def _deserialize_msg_body(self, fragment, offset, conn):
        error, offset = pdu.unpack_uint16(fragment, offset)
        self.error = tls.SSLError.val2enum(error)
        return self


"""Map the handshake message type to the corresponding class.
"""
_hs_deserialization_map = {
    tls.HandshakeType.HELLO_REQUEST: HelloRequest,
    tls.HandshakeType.CLIENT_HELLO: ClientHello,
    tls.HandshakeType.SERVER_HELLO: ServerHello,
    tls.HandshakeType.NEW_SESSION_TICKET: NewSessionTicket,
    tls.HandshakeType.END_OF_EARLY_DATA: EndOfEarlyData,
    tls.HandshakeType.ENCRYPTED_EXTENSIONS: EncryptedExtensions,
    tls.HandshakeType.CERTIFICATE: Certificate,
    tls.HandshakeType.SERVER_KEY_EXCHANGE: ServerKeyExchange,
    tls.HandshakeType.CERTIFICATE_REQUEST: CertificateRequest,
    tls.HandshakeType.SERVER_HELLO_DONE: ServerHelloDone,
    tls.HandshakeType.CERTIFICATE_VERIFY: CertificateVerify,
    # tls.HandshakeType.CLIENT_KEY_EXCHANGE = 16
    tls.HandshakeType.FINISHED: Finished,
    tls.HandshakeType.CERTIFICATE_STATUS: CertificateStatus,
    # tls.HandshakeType.KEY_UPDATE = 24
    # tls.HandshakeType.COMPRESSED_CERTIFICATE = 25
    # tls.HandshakeType.EKT_KEY = 26
    # tls.HandshakeType.MESSAGE_HASH = 254
}

_ccs_deserialization_map = {tls.CCSType.CHANGE_CIPHER_SPEC: ChangeCipherSpec}
"""Map the CCS message type to the corresponding class.
"""

_heartbeat_deserialization_map = {
    tls.HeartbeatType.HEARTBEAT_REQUEST: HeartbeatRequest,
    tls.HeartbeatType.HEARTBEAT_RESPONSE: HeartbeatResponse,
}

_ssl2_deserialization_map = {
    tls.SSLMessagType.SSL2_SERVER_HELLO: SSL2ServerHello,
    tls.SSLMessagType.SSL2_ERROR: SSL2Error,
}
"""Map the SSL2 message type to the corresponding class.
"""
